/*
 * $Id: CacheCore.java 286 2005-12-03 01:07:16Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.cache;

import java.util.*;

import org.apache.log4j.Logger;

/**
 * This is the core data structure for a DNS cache.
 * 
 * This is fairly agnostic about the data being cached or the keys for the
 * cache. It is up the the calling code to manage expiration times.
 * 
 * This data structure has a maximum size. Ideally, this would be expressed in
 * bytes. However, this being Java, knowing the size, in bytes, of an
 * arbitrary object is not an efficient operation. So currently, this
 * structure maintain a maximum number of entries (which misses some of the
 * logic, but what can we do?)
 * 
 * @author davidb
 * @version $Revision: 286 $
 */
public class CacheCore
{
  // NOTE: I felt it necessary to re-implement the double-linked list logic
  // rather than just using a LinkedList (or CircularFifoBuffer from
  // jakarta-commons-collections) because I need to be able to efficiently
  // *remove* an element found via the HashMap. The LinkedList class doesn't
  // give you the ability to do this, because it doesn't expose its internal
  // node, or really give you any way of doing this.

  /**
   * Cache entries. This contains the meta data associated with each cache
   * entry (i.e., the cached values are wrapped in these objects)
   */
  private static class Entry
  {
    /** The actual cached object. */
    Object value;
    /** The pointer to the next element in the LRU list. */
    Entry  next;
    /** The pointer to the previous element in the LRU list. */
    Entry  prev;

    public Entry(Object o)
    {
      this.value = o;
      this.next = this.prev = null;
    }
  }

  // This is the main lookup indexing device.
  private Map               mMap;
  // This is the LRU algorithm list header: next is the first element in the
  // list, prev is the last element in the list, object is always null.
  private Entry             mHeader;

  // The maximum number of objects allowed in the cache.
  private long              maxCacheSize;
  // The current number of objects in the cache.
  private long              cacheSize;
  // The default maximum number of objects allowed in the cache.
  private static final long DEFAULT_MAX_CACHE_SIZE = 1000000;

  private Logger            log                    = Logger.getLogger(this
                                                       .getClass());

  /**
   * Create an new CacheCore object. This will be initialized with a default
   * maximum cache size.
   */
  public CacheCore()
  {
    mMap = new HashMap();
    mHeader = new Entry(null);

    maxCacheSize = DEFAULT_MAX_CACHE_SIZE;

    // Initialize the list;
    mHeader.next = mHeader.prev = mHeader;
  }

  /**
   * Add an entry to the head of the Least Recently Used queue.
   * 
   * @param e The entry.
   */
  private void addToLRUHead(Entry e)
  {
    if (e == null) throw new NullPointerException();
    
    e.prev = mHeader;
    e.next = mHeader.next;
    mHeader.next = e;
    e.next.prev = e;

    cacheSize++;
  }

  /**
   * Remove an entry from the Least Recently Used queue, wherever it is in the
   * queue.
   * 
   * @param e The entry.
   */
  private void removeFromLRU(Entry e)
  {
    if (e == mHeader) throw new NoSuchElementException();

    if (e.prev == null || e.next == null)
      throw new IllegalArgumentException("Entry was invalid");

    e.prev.next = e.next;
    e.next.prev = e.prev;

    e.next = e.prev = null;

    cacheSize--;
  }

  /**
   * Delete the Least Recently Used object from the cache.
   * 
   * @return the deleted entry.
   */
  private Entry removeLRU()
  {
    Entry e = mHeader.prev;

    log.trace("removing Least Recently Used entry: " + e.value);
    removeFromLRU(e);

    return e;
  }

  /**
   * Remove all elements from the Least Recently Used queue.
   */
  private void clearLRU()
  {
    Entry next = mHeader.next;

    while (next != mHeader)
    {
      Entry e = next;
      next = e.next;

      e.prev = e.next = null;
    }
  }

  /**
   * Remove everything from the cache.
   */
  public synchronized void clear()
  {
    mMap.clear();
    clearLRU();
  }

  /**
   * Check to see if a particular key is present in the cache.
   * 
   * @param key The key.
   * @return true if the key is present in the cache.
   */
  public synchronized boolean containsKey(Object key)
  {
    return mMap.containsKey(key);
  }

  /**
   * Fetch an object from the cache.
   * 
   * @param key The key.
   * @return The object stored in the cache, if present. Otherwise, null.
   */
  public synchronized Object get(Object key)
  {
    Entry entry = (Entry) mMap.get(key);
    if (entry == null) return null;

    // Move this entry to the head of the queue.
    removeFromLRU(entry);
    addToLRUHead(entry);

    return entry.value;
  }

  /**
   * Put an object into the cache with the given key. This may be removed at
   * any time to make more room in the cache.
   * 
   * @param key The key.
   * @param value The object to put in the cache.
   * 
   * @return the value, as placed in the cache.
   */
  public synchronized Object put(Object key, Object value)
  {
    // Because we are measure size as number of entries in the map, this will
    // always execute once. If, at some point in the futre, measure size
    // differently, then this might loop more than one time.
    while (cacheSize >= maxCacheSize)
      removeLRU();

    Entry entry = new Entry(value);

    addToLRUHead(entry);

    mMap.put(key, entry);

    return entry.value;
  }

  /**
   * Explicitly remove an particular key from the cache.
   * 
   * @param key The key to remove.
   * @return the object that was removed, or null if the key did not exist in
   *         the cache.
   */
  public synchronized Object remove(Object key)
  {
    Entry entry = (Entry) mMap.remove(key);

    removeFromLRU(entry);

    return entry.value;
  }

  /**
   * Fetch the maximum cache size (in numbers of objects in the cache, not
   * bytes)
   * 
   * @return the size.
   */
  public long getMaxCacheSize()
  {
    return maxCacheSize;
  }

  /**
   * Set the maxium cache size (in numbers of objects in the cache, not
   * bytes).
   * 
   * @param maxCacheSize the size.
   */
  public void setMaxCacheSize(long maxCacheSize)
  {
    this.maxCacheSize = maxCacheSize;
  }
}